/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * Created on : 04.2003
 * Created by : s3595
 */
package at.spardat.xma.boot.cache;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.SimpleTimeZone;

import at.spardat.xma.boot.Statics;
import at.spardat.xma.boot.logger.LogLevel;
import at.spardat.xma.boot.logger.Logger;
import at.spardat.xma.boot.util.PropertyFile;
import at.spardat.xma.boot.util.Util;


/**
 *
 * class: FCResource ( File Cached Resource ) <p>
 *
 * Describes an entry in the file-cache.<p>
 *
 * @author s3595 <a href="mailto:schaeferc@spardat.at">Chris Schfer (CGS)</a>
 * @version $Id: FCResource.java 2084 2007-11-27 14:53:31Z s3460 $
 */
public class FCResource implements IFileCacheResource {

    /** the remote resource location */
    protected URL location_;

    /** info-file about the cached file */
    protected PropertyFile properties_;

    /** resource file */
    protected File file_;

    /** byte array for buffered mode */
    protected byte[] buffer_;

    /** extended readable file output */
    protected boolean debug;

    /** used to format timesamps for debug correctly (converted long milliseconds since 1.1.1970 00:00:00 GMT) */
    private DateFormat formater;

    VersionNumber versionNumber;

    /**
     * <p> Create a <tt>FCResource</tt> for the resources specified as a remote URL.</p>
     *
     * <p> it is protected, because a resource must not be created by others
     * than the FileCache </p>
     *
     * <p>If the file does not exist, it will not be created.<p>
     *
     * @param   location   the remote resource location
     * @param   file       the resource file
     * @param   bmode      buffer mode
     * @throws IOException containing the filename
     * @throws   IllegalArgumentException   the file must exist already
     */
    protected FCResource(URL location, File file, boolean bmode, boolean debug) throws IOException {
        File infoFile = null;

//        if(location == null || file == null ) {
//          throw new IllegalArgumentException( Statics.strInvalidARG );
//        }
        if( !file.exists() ) {
            throw new FileNotFoundException( "file: " + file.getAbsolutePath()); //$NON-NLS-1$
        }

        formater = new SimpleDateFormat( "EEE, dd MMM yyyy hh:mm:ss zzz", Locale.getDefault() );            //$NON-NLS-1$
        formater.setTimeZone(new SimpleTimeZone(0, "GMT")); //$NON-NLS-1$

        this.debug       = debug;
        this.location_   = location;
        this.file_       = file;
        this.buffer_     = null;
        infoFile         = new File( file.getPath() + Statics.FILEINFO_EXT);
        this.properties_ = new PropertyFile(infoFile, debug);

        if(bmode) {
            loadBuffer();
        }

    }// FCResource

    /**
     * load data from resource file into memory buffer
     *
     * @return void
     * @throws IOException containing the filename
     */
    protected void loadBuffer() throws IOException {

        int length     = (int)file_.length();
        buffer_        = new byte[length];
        InputStream is = null;
        int read       = 0;

        try {
            is = new FileInputStream( file_ );
            for(int all=0; all<length ; all+=read ) {
                read = is.read( buffer_, all, length-all );
                if(read<0) {  break; } // error in reading
            }
        } catch (IOException exc) {
            throw new IOException("Error reading resource-file "+file_.getAbsolutePath());
        } finally {
            Util.close(is,file_.getAbsolutePath());
        }
        if(read<0) { // loop terminated by error
            throw new IOException("Error reading resource-file "+file_.getAbsolutePath());
        }
    }

    /**
     * Returns the remote location this entry caches.
     */
    public URL getLocation() {
        return location_;
    }

    /**
     * Saves the current information for the cache entry.
     * @throws IOException containing the filename
     */
    protected void store() throws IOException {
        properties_.store();
    }

    /**
     * @see IFileCacheResource
     * @return InputStream
     * @throws IOException containing the filename
     */
    public InputStream getInputStream() throws IOException {
        if(buffer_== null) {
           loadBuffer();
        }
        return new ByteArrayInputStream(buffer_);
    }

    /**
     * for internal use in file cache only!
     * @throws IOException containing the filename
     */
    byte[] getBuffer() throws IOException {
        if(buffer_==null) loadBuffer();
        return buffer_;
    }

    /**
     * Returns the value of the specified key, or null if the key
     * does not exist.
     *
     * @param key key to look for
     */
    public String getProperty(String key) {
      return properties_.getProperty(key);
    }

    /**
     * Returns the value of the specified key, or the default value
     * if the key does not exist.
     */
    public String getProperty(String key, String defaultValue) {
      return properties_.getProperty(key, defaultValue);
    }

    /**
     * Sets the value for the specified key.
     * @return the previous value
     */
    public Object setProperty(String key, String value) {
        return this.properties_.setProperty(key,value);
    }

    /**
     * Returns the time in the local system clock that the file was
     * most recently checked for an update.
     */
    public long getLastUpdated() {
        try {
            String strLastUpdated = properties_.getProperty( Statics.strLastUpdated);
            if(strLastUpdated!=null) return Long.parseLong(strLastUpdated);
            else return 0;
        } catch (Exception ex) {
            Logger.getLogger("bootrt.bootRuntime").log(LogLevel.WARNING,
                "error parsing "+Statics.strLastUpdated+" in "+properties_.getStoreFile().getAbsolutePath(),ex);
            return 0;
        }
    }

    /**
     * Sets the time in the local system clock that the file was
     * most recently checked for an update.
     */
    public void setLastUpdated(long updatedTime) {
        properties_.setProperty(Statics.strLastUpdated, Long.toString(updatedTime));
        if(debug) {
            properties_.setProperty(Statics.strLastUpdated + ".STR", formater.format(new Date(updatedTime)) );
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.cdb.cache.IFileCacheResource#getLastModified()
     */
    public long getLastModified() {
        try {
            String strLastModified = properties_.getProperty( Statics.HTTP_LAST_MODIFIED);
            if(strLastModified!=null) return Long.parseLong(strLastModified);
            else return 0;
        } catch (Exception ex) {
            Logger.getLogger("bootrt.bootRuntime").log(LogLevel.WARNING,
                    "error parsing "+Statics.HTTP_LAST_MODIFIED+" in "+properties_.getStoreFile().getAbsolutePath(),ex);
            return 0;
        }
    }

    /**
     * Sets the value of the last modified header in the resource description file.
     * @param lastModified in milliseconds since 1.1.1970
     */
    public void setLastModified(long lastModified) {
        properties_.setProperty(Statics.HTTP_LAST_MODIFIED, Long.toString(lastModified));
        if(debug) {
            properties_.setProperty(Statics.HTTP_LAST_MODIFIED + ".STR", formater.format(new Date(lastModified)) );
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.cdb.cache.IFileCacheResource#setExpiration(long)
     */
    public void setExpiration(long lExpires ) {
        properties_.setProperty(Statics.HTTP_EXPIRES, Long.toString(lExpires));
        if(debug) {
            properties_.setProperty(Statics.HTTP_EXPIRES + ".STR", formater.format(new Date(lExpires)) );
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.cdb.cache.IFileCacheResource#getExpiration()
     */
    public long getExpiration() {
        try {
            String strExpires = properties_.getProperty( Statics.HTTP_EXPIRES);
            if(strExpires!=null) return Long.parseLong(strExpires);
            else return 0;
        } catch (Exception ex) {
            Logger.getLogger("bootrt.bootRuntime").log(LogLevel.WARNING,
                    "error parsing "+Statics.HTTP_EXPIRES+" in "+properties_.getStoreFile().getAbsolutePath(),ex);
            return 0;
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.cdb.cache.IFileCacheResource#setProperty(java.lang.String, long)
     */
    public void setProperty(String string, long l) {
        this.setProperty(string, Long.toString(l));
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.cdb.cache.IFileCacheResource#getLocalRes()
     */
    public URL getLocalRes() {
        if( file_ == null ) {
          return null;
        }

        try {
            return file_.toURL();
        } catch (MalformedURLException e) {
            throw new RuntimeException("error converting "+file_.getAbsolutePath()+" to url",e);
        }
    }

    /*
     *  (non-Javadoc)
     * @see at.spardat.xma.boot.cache.IFileCacheResource#isExpired()
     */
    public boolean isExpired() {

        long expires = this.getExpiration();
        long ctime   = System.currentTimeMillis();

        // if the resource has exiriration information set,
        // and it is exired.
        if( expires != 0 && expires < ctime )
          return true;
        else
          return false;
    }

     File getResourceFile() {
        return file_;
     }

     File getProperties() {
        return properties_.getStoreFile();
     }

}


